#include "pyggi.h"
#include "pyft2.h"

pyggi_C_api * ggi_c_api;
pyft2_C_api * ft2_c_api;

static inline void
blend_color(ggi_color *dst, ggi_color *c1, ggi_color *c2) {
	dst->r=c1->r+(((c2->r-c1->r)*(c2->a >> 8)) >> 8);
	dst->g=c1->g+(((c2->g-c1->g)*(c2->a >> 8)) >> 8);
	dst->b=c1->b+(((c2->b-c1->b)*(c2->a >> 8)) >> 8);
}

/* This is extremely slow.... */
static void
ggiFTBitmapBlitAlpha(ggi_visual_t vis, int x, int y, ggi_color *col,
		     FT_Bitmap * bitmap,
		     int bbox[4]) {
	int i,j;
	int sx, sy;
	int dx, dy;
	ggi_color c;
	ggi_color c2;
	ggi_pixel p;
	unsigned char * buf;
	int alpha;
	buf   = bitmap->buffer;
	c     = *col;
	alpha = c.a >> 8;
	
	/* compute the visible rectangle */
	
	sx = sy = 0;
	dx = bitmap->width;
	dy = bitmap->rows;

	if ( (x > bbox[2]) || (y>bbox[3]) ||
	     ((x+dx) <= bbox[0]) || ((y+dy) <= bbox[1]))
		return;
	
	if (bbox[0] > x)    sx = bbox[0]-x;
	if (bbox[1] > y)    sy = bbox[1]-y;
	if (bbox[2] < x+dx) dx = x - bbox[2] + 1;
	if (bbox[3] < y+dy) dy = y - bbox[3] + 1;

	/* printf("%i,%i %ix%i -> %i,%i\n",sx,sy,dx,dy,x+sx,y+sy); */
	
	buf += bitmap->width*sy + sx;
	
	for(j=sy;j<dy;j++) {
		for(i=sx;i<dx;i++)
			if (buf[i]) {
				ggiGetPixel(vis,x+i,y+j,&p);
				ggiUnmapPixel(vis,p,&c2);
				c.a = alpha *(buf[i]);
				/* printf ("%i -> %i\n",*buf, c.a); */
				blend_color(&c2,&c2,&c);
				ggiPutPixel(vis,x+i,y+j,ggiMapColor(vis,&c2));
			}
		buf += bitmap->pitch;	      
	}
}

static PyObject *
ggift2_putGlyph(PyObject *self, PyObject *args, PyObject *kwds) {
	PyObject * pf;
	PyObject * vis;
	
	int bbox[4];
	int x,y;
	float r,g,b,a;
	FT_Face f;
	ggi_visual_t v;
	ggi_color c;
	
	if(!PyArg_ParseTuple(args, "OOii(ffff)", &vis, &pf, &x, &y,
			     &r, &g, &b,&a ))
		return NULL;

	ASSERT_VISUAL(vis, NULL);
	v = ((Visual*)(vis))->vis;
	ASSERT_FACE(pf, NULL);
	f = ((Face*)(pf))->face;
	float2color(r,g,b,&c);
	c.a = (unsigned int)(0xffff*a);
	
	ggiGetGCClipping(v,&bbox[0],&bbox[1],&bbox[2],&bbox[3]);
	ggiFTBitmapBlitAlpha(v, x+f->glyph->bitmap_left, 
			     y - f->glyph->bitmap_top,
			     &c,&f->glyph->bitmap, bbox);
	Py_INCREF(Py_None);
	return Py_None;
}

static const char ggift2_write_doc[] = "not yet";
static PyObject *
ggift2_putGlyphs(PyObject *self, PyObject *args, PyObject *kwds) {
	PyObject * pf;
	PyObject * vis;
	int x,y;
	char * s;
	float r,g,b,a;
	FT_Face f;
	ggi_visual_t v;
	ggi_color c;
	
	int px, py;
	int xerr;
	int gidx;
	int bbox[4];
	
	/* args: Face, x, y, str, [ color ]*/
	if(!PyArg_ParseTuple(args, "OOiis(ffff)", &vis, &pf, &x, &y, &s,
			     &r, &g, &b,&a )) {
		return NULL;
	}
	/* option : kerning, atialias */

	ASSERT_VISUAL(vis, NULL);
	v=((Visual*)(vis))->vis;
	ASSERT_FACE(pf, NULL);
	f=((Face*)(pf))->face;
	
	float2color(r,g,b,&c);
	c.a=(unsigned int)(0xffff*a);

	px   = x;
	py   = y;
	xerr = 0;
	ggiGetGCClipping(v,&bbox[0],&bbox[1],&bbox[2],&bbox[3]);
	while(*s) {
		gidx = FT_Get_Char_Index( f, (FT_ULong)*s);
		if(!FT_Load_Glyph(f, (FT_UInt)gidx, FT_LOAD_RENDER)) {
			/* we don't do kerning */
			ggiFTBitmapBlitAlpha(v, px+f->glyph->bitmap_left, 
					     py - f->glyph->bitmap_top,
					     &c,&f->glyph->bitmap,bbox);
			xerr += f->glyph->linearHoriAdvance & 0xff;	
			px   += xerr >> 16;
			xerr &= 0xff;
			px   += f->glyph->linearHoriAdvance >> 16;
			/*pen_y += f->glyph->linearVertAdvace >> 16;*/
		}
		s++;
	}
	/* return the output bbox? */
	Py_INCREF(Py_None);
	return Py_None;
}


static struct PyMethodDef ggift2module_methods[] = {
	{ "putGlyphs", (PyCFunction)ggift2_putGlyphs, METH_VARARGS,
	  (char*)NULL },
	{ "putGlyph",  (PyCFunction)ggift2_putGlyph,  METH_VARARGS,
	  (char*)NULL },
	{ NULL, NULL, NULL, NULL }
};


static void importModules(void) {
	PyObject *ggi;
	PyObject *ft2;
	
	PyObject *ggidict;
	PyObject *ft2dict;
	PyObject *capiobj;
	
	/* import ggi */
	
	ggi = PyImport_ImportModule("GGI._ggimodule");
	if (ggi != NULL) {
		ggidict = PyModule_GetDict(ggi);
		capiobj = PyDict_GetItemString(ggidict, "_ggimodule_C_api");
		if (PyCObject_Check(capiobj)) {
			ggi_c_api=(pyggi_C_api *)PyCObject_AsVoidPtr(capiobj);
		}
		/* 
		Py_DECREF(capiobj);
		*/
	} else {
		Py_FatalError("Can't import GGI._ggimodule");		
		return;
	}
	
	/* import ft2 */
	ft2 = PyImport_ImportModule("GGI._ft2module");
	if (ft2 != NULL) {
		ft2dict = PyModule_GetDict(ft2);
		capiobj = PyDict_GetItemString(ft2dict, "_ft2module_C_api");
		if (PyCObject_Check(capiobj)) {
			ft2_c_api=(pyft2_C_api *)PyCObject_AsVoidPtr(capiobj);
		}
		/*
		Py_DECREF(capiobj);
		*/
	} else {
		Py_FatalError("Can't import GGI._ft2module");
		return;
	}
}


#ifndef PyMODINIT_FUNC
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
init_ggift2module(void) {
	
	importModules();
	Py_InitModule3("_ggift2module", ggift2module_methods, (char *)NULL);
	
	if(PyErr_Occurred())
		Py_FatalError("Something went wrong in GGI._ggift2module ");
}
